"""
constants.py contains constants used throughout the application.
"""
#: The application name for Google Content.
APPLICATION_NAME = "Content API for Shopping"

# The product details are provided in English.
CONTENT_LANGUAGE = "en"

# The products are sold in the USA.
TARGET_COUNTRY = "US"

# The products will be sold online.
CHANNEL = "online"

#: Service account API key file.
#: If credentials are not passed in to the service,
#: it will default to the file at this path.
SERVICE_ACCOUNT_FILE = "../credentials/charged-curve-410406-b9ed7ae95369.json"

#: The API URI.
API_URI = "https://www.googleapis.com/discovery/v1/apis/content/v2.1/rest"

#: The content scope.
CONTENT_API_SCOPE = "https://www.googleapis.com/auth/content"

#: The service name.
SERVICE_NAME = "content"

#: The service version.
SERVICE_VERSION = "v2.1"

#: The sandbox service version.
SANDBOX_SERVICE_VERSION = "v2.1sandbox"

#: The merchant ID.
# If your merchant ID is not passed in to the service,
# it will default to this value.
MERCHANT_ID = "5321244390"

#: The maximum page size.
MAX_PAGE_SIZE = 50

#: The default batch size.
#: Google's maximum is 1000
#: https://developers.google.com/shopping-content/v2/how-tos/batch
BATCH_SIZE = 1000

# Product Availability - In Stock
AVAILABILITY_IN_STOCK = "in stock"

# Product Availability - Out of Stock
AVAILABILITY_OUT_OF_STOCK = "out of stock"

# Product Availability - Preorder
AVAILABILITY_PREORDER = "preorder"



"""
utils.py contains utility functions used throughout the application.
"""
import logging

import gshoppy.constants as constants
from google.oauth2 import service_account


def setup_logging() -> None:
    """
    Sets up logging.
    :return:
    """

    # Set up basic config.
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )


def get_logger() -> logging.Logger:
    """
    Gets the logger.
    :return:
    """

    return logging.getLogger()


def get_product_id(offer_id: str) -> str:
    """
    Get the product ID from the offer ID.
    :param offer_id: The offer ID.
    :return: The product ID.
    """

    return f"{constants.CHANNEL}:{constants.CONTENT_LANGUAGE}:{constants.TARGET_COUNTRY}:{offer_id}"


def get_credentials_from_file(path: str) -> service_account.Credentials:
    """
    Loads credentials from a file.
    :param path:
    :return:
    """

    return service_account.Credentials.from_service_account_file(
        path, scopes=[constants.CONTENT_API_SCOPE]
    )



"""
models.py contains the models for the application.
"""
from enum import Enum
from typing import Final

import gshoppy.constants as constants


class ProductAvailability(str, Enum):
    """
    The product availability.
    """

    IN_STOCK: Final[str] = constants.AVAILABILITY_IN_STOCK
    OUT_OF_STOCK: Final[str] = constants.AVAILABILITY_OUT_OF_STOCK
    PREORDER: Final[str] = constants.AVAILABILITY_PREORDER



"""
product.py contains the Product service class.
"""
from abc import ABC
from typing import Generator
from typing import List

import gshoppy.constants as constants
from gshoppy.services.core import Service
from gshoppy.utils import get_logger
from gshoppy.utils import get_product_id
from gshoppy.utils import setup_logging

setup_logging()
logger = get_logger()


class ProductService(Service, ABC):
    """
    The Product service class.
    """

    def _init_service(self):
        """
        Initializes the Product Service.
        :return:
        """

        logger.info("Initializing the Product service.")

        self._service = self.base_service.products()

    def get(self, offer_id: str) -> dict:
        """
        Gets a product.
        :param offer_id: The offer ID of the product to get.
        :return:
        """

        logger.info("Getting product.")

        # Get the product ID.
        product_id = get_product_id(offer_id)

        # Build a request.
        request = self.service.get(merchantId=self.merchant_id, productId=product_id)

        # Execute the request.
        return request.execute()

    def list(self) -> Generator:
        """
        Yield out pages of products.
        :return:
        """

        logger.info("Fetching products")

        # Build a request.
        request = self.service.list(
            merchantId=self.merchant_id, maxResults=constants.MAX_PAGE_SIZE
        )

        # Get all pages.
        while request is not None:
            # Execute the request.
            response = request.execute()

            # Yield the products.
            yield response.get("resources")

            # Update the request.
            request = self.service.list_next(
                previous_request=request, previous_response=response
            )

    def insert(self, **product) -> dict:
        """
        Inserts a product.
        :param product: The product to insert.
        :return:
        """

        logger.info("Inserting product.")

        # Build a request.
        request = self.service.insert(merchantId=self.merchant_id, body=product)

        # Execute the request.
        return request.execute()

    def update(self, offer_id: str, **product) -> dict:
        """
        Updates a product.
        :param offer_id:
        :param product:
        :return:
        """

        logger.info("Updating product.")

        # Get the product.
        _product = self.get(offer_id)

        # Update the product.
        _product.update(product)

        # Cannot specify `source` when inserting a product.
        del _product["source"]

        # Re-insert the product.
        return self.insert(**_product)

    def delete(self, offer_id: str) -> None:
        """
        Deletes a product.
        :param offer_id: The offer ID of the product to delete.
        :return:
        """

        logger.info("Deleting product.")

        # Get the product ID.
        product_id = get_product_id(offer_id)

        # Build a request.
        request = self.service.delete(merchantId=self.merchant_id, productId=product_id)

        # Execute the request.
        request.execute()

    def batch_insert(
        self, *products: List[dict], batch_size: int = constants.BATCH_SIZE
    ):
        """
        Inserts products in batches.
        :param products:
        :param batch_size:
        :return:
        """

        raise NotImplementedError()

    def batch_delete(
        self, *offer_ids: List[str], batch_size: int = constants.BATCH_SIZE
    ):
        """
        Deletes products in batches.
        :param offer_ids:
        :param batch_size:
        :return:
        """

        raise NotImplementedError()



"""
service.py contains the Service class, which is the base class for all services.
"""
from abc import ABC
from abc import abstractmethod
from typing import Generator
from typing import Sized

import google_auth_httplib2
import gshoppy.constants as constants
from google.oauth2 import service_account
from googleapiclient import discovery
from googleapiclient import http
from gshoppy.utils import get_logger
from gshoppy.utils import setup_logging

setup_logging()
logger = get_logger()


class Service(ABC):
    """
    The base class for all services.
    """

    def __init__(
        self,
        credentials: service_account.Credentials = None,
        merchant_id: str = None,
        sandbox: bool = False,
    ):
        """
        Initializes the service.

        :param sandbox: Whether to use the sandbox API.
        """

        self._sandbox = sandbox

        self._credentials = credentials
        self._merchant_id = merchant_id
        self._base_service = None

        # Initialize the service.
        self._service = None
        self._init_service()

    @property
    def sandbox(self) -> bool:
        """
        Whether this service uses the sandbox API.
        :return: Whether to use the sandbox API.
        """

        return self._sandbox

    @property
    def credentials(self) -> service_account.Credentials:
        """
        The service account credentials.
        :return:
        """

        if self._credentials is None:
            logger.warning(
                f"No credentials provided - defaulting to {constants.SERVICE_ACCOUNT_FILE}"
            )
            self._credentials = service_account.Credentials.from_service_account_file(
                constants.SERVICE_ACCOUNT_FILE, scopes=[constants.CONTENT_API_SCOPE]
            )

        return self._credentials

    @property
    def merchant_id(self) -> str:
        """
        The merchant ID.
        :return: The merchant ID.
        """

        if self._merchant_id is None:
            logger.warning(
                f"No merchant ID provided - defaulting to {constants.MERCHANT_ID}"
            )
            self._merchant_id = constants.MERCHANT_ID

        return self._merchant_id

    @property
    def base_service(self) -> discovery.Resource:
        """
        The base service.
        :return: The service.
        """

        if self._base_service is None:
            logger.info("Initializing the base service.")

            # Get the Auth HTTP for the service.
            http_arg = http.set_user_agent(
                http.build_http(), constants.APPLICATION_NAME
            )
            auth_http = google_auth_httplib2.AuthorizedHttp(
                self.credentials, http=http_arg
            )

            # Set service.
            self._base_service = discovery.build(
                constants.SERVICE_NAME,
                self.version,
                discoveryServiceUrl=constants.API_URI,
                http=auth_http,
            )

        return self._base_service

    @property
    def version(self) -> str:
        """
        The service version.
        :return: The service version.
        """

        return (
            constants.SANDBOX_SERVICE_VERSION
            if self.sandbox
            else constants.SERVICE_VERSION
        )

    @property
    def service(self) -> discovery.Resource:
        """
        The service.
        :return:
        """

        return self._service

    @staticmethod
    def _batch(iterable: Sized, size: int = constants.BATCH_SIZE) -> Generator:
        """
        Batches an iterable into chunks of a given size.
        :param iterable: The iterable to batch.
        :param size: The batch size.
        :return: The batched iterable.
        """

        for i in range(0, len(iterable), size):
            yield iterable[i : i + size]

    @abstractmethod
    def _init_service(self):
        """
        Initializes the service.
        :return:
        """

    def __del__(self):
        """
        Close any open service sockets.
        :return:
        """

        if self.service is not None:
            self.service.close()
        if self.base_service is not None:
            self.base_service.close()



from setuptools import setup, find_packages

VERSION = '0.0.7'
DESCRIPTION = 'A package to facilitate the use of Google Shopping API'

setup(
    name="gshoppy",
    version=VERSION,
    author="Jake Williamson",
    author_email="<brianjw88@gmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=['requests', 'pandas', 'google-api-python-client', 'google-auth-httplib2', 'google-auth-oauthlib'],
    keywords=['python', 'google', 'shopping', 'api', 'google shopping api'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3"
    ]
)
